local class_mt = {}

function class_mt:__index(key)
    return self.__baseclass[key]
end

function class_mt:__call(...)
  return self:new(true, ...)
end

--@param(k : string)
function class_mt:__newindex(k,v)
  if k:sub(1,2) == '__' then
    rawset(getmetatable(self), k, v)
  else
    rawset(self,k,v)
  end
end

local function makeSubscriptionInterfaces(instEvents, classEvents)
  for k,v in pairs(classEvents) do
    instEvents[k] = { subscribers = {} }
    
    local function subscribe(self,subscriber)
        local subscribers = self.subscribers
        subscribers[#subscribers+1] = subscriber
    end
    
    instEvents[k].subscribe = subscribe
  end
end

local function addEventTriggers(inst)
  local events = inst.events
  
  for k,v in pairs(events) do
    inst[k] = function(self,...)
      for _,subscriber in ipairs(v.subscribers) do
        subscriber(...)
      end
    end
  end
end

class = setmetatable({ __baseclass = {} }, class_mt)

--TODO: I need to separate this into two functions
function class:new(inst,...)
  --@var(unknown)
  local c = {}

  --@var(unknown)
  local baseInit
  if self.__baseclass then
    baseInit = self.__baseclass.init
  end

  if not inst then
    c.events = {}
  else
    c.events = {}
    makeSubscriptionInterfaces(c.events,self.events)
    addEventTriggers(c)
  end
  
  c.__baseclass = self
  setmetatable(c, getmetatable(self))
  if c.init then
      c.init_super = baseInit
      c:init(...)
      c.init_super = nil
  end
  return c
end 

return class